// ----------------------------- SMSTCPGateway -----------------------------------------------------------------------------------------------
//
// This sketch is designed to work with the Haute Solutions ESP324G Board with a Haute Solutions OLED Panel
// (www.haute-solutions.com)
//
// Haute Solutions Libraries Required: SMSDEV ( GSMSMS, OLED_SSD1311 )
// External Libraries Required:  Rob Tillart's PCF8574 Library: https://github.com/RobTillaart/PCF8574
//
// This sketch will listen for a TCP connection on the specified port and then convert and send that message
// as an SMS text message to the designated SMS Phone number.  Inbound message format on TCP port should be
// multiple lines of text formatted as follows:
//
//      {SenderIP}
//      SMS_SEND
//      {SMS Recipient Number}
//      {Message Line1}   
//      {Message Line2}      
//      {Message Line...}   
//
//      For Example:
//
//      192.168.0.1
//      SMS_SEND
//      +12345678900
//      This is a multiline
//      message!
//
// This sketch will also listen for inbound SMS messages which will then be sent to the specified port on a
// Server for additional processing/handling.  SMS Messages will be reformatted as follows:
//
//      {DeviceIP}
//      SMS_RCV
//      {SMS Sender Number}
//      {Message Line1}   
//      {Message Line2}      
//      {Message Line...}   
//
//      For Example:
//
//      192.168.0.2
//      SMS_RCV
//      +12345678900
//      This is a multiline
//      message!
//
// IP Address and Port of Server to handle/process inbound SMS messages must be specified.
//
// OLED/Button I2C Connections (Deafults):
//                          GND to GND
//                          VCC to 3V3
//                          SDA to SDA (IO21)
//                          SCL to SCL (IO22)
//
// This file is free software; you can redistribute it and/or modify
// it under the terms of either the GNU General Public License version 2
// or the GNU Lesser General Public License version 2.1, both as
// published by the Free Software Foundation.
//
// --------------------------------------------------------------------------------------------------------------------------
#include "SMSDEV.h"
#include "PCF8574.h"
#include <WiFi.h>

#define COMPANY   "Haute Solutions"
#define DEVICE    "WiGGBox"
#define VERSION   "V1.0"
#define EEPROMSIG "SMSWiFiGateway 1.0"    // EEPROM Valid Signature (19 Char Max)

#define BOARD ESP324G                     // Compile for ESP324G
#define OLEDADDR 0x3C                     // I2C Address for OLED Display (NULL if None)
#define PCFADDR 0x38                      // I2C Address for PCF8574 Button Pad
#define BUTTON1_PIN 13                    // Display Main Screen           
#define BUTTON2_PIN 32                    // Display NetInfo
#define BUTTON3_PIN 33                    // Display CfgInfo
#define BUTTON4_PIN 17                    // Factory Reset
// SDA = 21
// SCL = 22

CMDINFO Commands[] = {                    // Device Specific Commands (Generic SMS Commands are defined in SMSDEV Library)
  {"HELP",         1, 0},
  {"STATUS",       2, 0},
  {"WIFISSID:",    3, 1},
  {"WIFIPWD:",     4, 1},
  {"SERVERIP:",    5, 1},
  {"NOTIFY:",      6, 1}
};

//--------------------------------------------------------------------------------------------------------------------------------------
//  The following global parameters are set (initially) in the Site_Data.h file and 
//  maintained in EEPROM so they are persistant and can be remotely updated.
char WiFiSSID[EEFIELDSIZE];          // WiFi SSID
char WiFiPwd[EEFIELDSIZE];           // WiFi Password
char ServerIP[EEFIELDSIZE];          // IP Address of Server for processing SMS Messages                                                     
char Notify[EEFIELDSIZE];            // SMS Recipient for Heartbeat Notification
char Password[EEFIELDSIZE];          // Password to authorize a Device
//--------------------------------------------------------------------------------------------------------------------------------------

// Define some global shorthand "aliases" (for readability)
char* Buffer;
char* Caller;
char* Temp;
char* Param;
char* DevName;
char* DevNum;

// IP Info obtained via WiFi
char WiGGIP[16];                      // IP Address for the WiGGBox
char WiGGMAC[18];                     // MAC Address for the WiGGBox

SMSDEV SMSDev(BOARD, OLEDADDR);
PCF8574 PCF(PCFADDR);

//#include "Site_Data.h"
#include "Site_Data_ECV.h"

// Inbound TCP port to listen on
// Currently, Ethernet Library does not allow us to set this later (i.e. Setup())
WiFiServer TCPServer(ServicePort);
WiFiClient client;

//  The following parameters are loaded from EEPROM or Site_Data.h as appropriate
IPAddress ServerIPAddr( 0, 0, 0, 0 );                         // IP Address of the Server to process SMS Messages

#define TCPBUFFSIZE 512
char TCPBuffer[TCPBUFFSIZE];

int NetAvail;

void setup(){
int rcode;

  Serial.begin(9600);

  // Uncomment the line below to Wack the EEPROM Signature and force a factory reset
  // EEPROM.write(0, '\0');
  
  SetButtons();

  rcode = SMSDev.Setup(COMPANY, DEVICE, VERSION, EEPROMSIG, "#", Commands, sizeof(Commands)/sizeof(Commands[0]));
  while (rcode > 0) { }
  
  SMSDev.SetPrivProcessing(0);      // Disable Priv Command Checking

  // Assign our Global Aliases
  Buffer       = SMSDev.Buffer;
  Caller       = SMSDev.Caller;
  Temp         = SMSDev.Temp;
  Param        = SMSDev.Param;
  DevName      = SMSDev.DevName;
  DevNum       = SMSDev.DevNum;

  if(!GetDefaults()) SetDefaults();
  
  if(StartNet() != 0) NetAvail = 0;
  else NetAvail = 1;

  if(NetAvail) {
    IPFromString(ServerIP, &ServerIPAddr);
    SMSDev.DisplayWithDelay("Server Address", ServerIP, 1);
    SMSDev.DisplayWithDelay("Service Port", itoa(ServicePort, Temp, 10), 1);
  }

  if(SMSDev.GetAuthCaller(0, Caller) != 0) {
    if(SMSDev.BootNotify[0] == '1') {
      CommandSTATUS(Caller, "System Ready!");
    }
  }

  SendTCPMessage("HELLO", "WIGGBOX", NULL);

  //Process any pending/stored SMS Messages
  while (SMSDev.CheckSMS() == 1) {
  }

  SMSDev.DisplayIdleScreen();
}

void loop(){
    
  switch (SMSDev.Loop()) {
    case -7:  // CFGINFO (SMSDEV) = Add additional STATUS Screen
      CommandSTATUS(Caller, NULL);
      break;
    case -9:  // Factory Reset (SMSDEV)
      CommandRESET();
      break;

    case 1:   // HELP
      CommandHELP();
      break;
    case 2:   // STATUS
      CommandSTATUS(Caller, NULL);
      break;
    case 3:   // WIFISSID:
      SetConfig(Param, WiFiSSID, EEDEVSPEC1, "WiFi SSID Set! Rebooting...", Caller);
      SMSDev.Reboot();
      break;
    case 4:   // WIFIPWD:
      SetConfig(Param, WiFiPwd, EEDEVSPEC2, "WiFi Password Set! Rebooting...", Caller);
      SMSDev.Reboot();
      break;
    case 5:   // SERVERIP:
      SetConfig(Param, ServerIP, EEDEVSPEC3, "Server IP Set! Rebooting...", Caller);
      SMSDev.Reboot();
      break;
    case 6:   // NOTIFY:
      SetConfig(Param, Notify, EEAUTHCALLERS, "Notification Number Set!", Caller);
      break;

    case 100: // Clock Tick (every 30 sec)
      break;
    case 101: // Alarm 1
      break;
    case 102: // Alarm 2
      break;
    case 103: // HeartBeat
      if(SMSDev.GetAuthCaller(0, Caller) != 0) CommandSTATUS(Caller, "HeartBeat!");
      break;
    case 104: // CSQ (Signal Change)
      // Display Updated in SMSDEV
      break;
    case 200: // Unauthorized Caller 
      // Message Forwarded in SMSDEV
      break;
    case 201: // Invalid Command
      // Reply Handled in SMSDEV
      break;
    case 202: // Command Restricted
      // Reply Handled in SMSDEV
      break;
    case 203: // Forwarded Message for local Processing (Not a Command)
      SendTCPMessage("SMS_RCV", Caller, Buffer);
      break;
  }

  CheckTCP();                                 // Any TCP Messages Pending?

  CheckButtons();
}

// ------------------------------ TCP Support Functions -----------------------------------------

int StartNet() {
int i;  
byte MAC[6];

  if(strlen(WiFiSSID) == 0) {
    SMSDev.DisplayWithDelay("Use WIFISSID", "Command to Set!", 1);
    return 1;
  }

  SMSDev.DisplayWithDelay("Starting", "WiFi", 1);

  WiFi.begin(WiFiSSID, WiFiPwd);
 
  WiFi.macAddress(MAC);
  MACToString(MAC, WiGGMAC);
  SMSDev.DisplayWithDelay("MAC Address:", WiGGMAC, 1);

  for(i=0; i<20; i++) {
    if(WiFi.status() == WL_CONNECTED) break;
    delay(500);
  }
  if(i == 20) {
      SMSDev.DisplayWithDelay("WiFi Fail!", WiFiSSID, 1);
      return 2;
  }
  SMSDev.DisplayWithDelay("Connected to:", WiFiSSID, 1);

  strcpy(WiGGIP,WiFi.localIP().toString().c_str());
  SMSDev.DisplayWithDelay("IP Address:", WiGGIP, 1);

  return 0;
}

void CheckTCP() {  
char c;
int offset;
unsigned long timeout;

WiFiClient client;

  client = TCPServer.available();
  if (!client) return;

  TCPBuffer[0] = '\0';
  offset=0;
    
  while (client.connected()) {
    if (client.available()) {
        
      c = client.read();
      timeout = millis() + 1000;
      if(offset < (TCPBUFFSIZE-2)) {
        TCPBuffer[offset++] = c;
        TCPBuffer[offset] ='\0';
      }
      
    } else {
      if(millis() > timeout) {
        client.println("OK");
        break;
      }
    }   
  }

  client.stop();
  ProcessTCPMessage(TCPBuffer);
  Serial.println();
}

int ProcessTCPMessage(char *Cmd) {
char *p1;
char *p2;
char *p3;
char *p4;
char *p5;

  Serial.println("TCP Incoming:");
  Serial.println(Cmd);
  Serial.println("");

  p1=Cmd;
  p2=NULL;
  p3=NULL;
  p4=NULL;

  p2=strchr(p1,'\n');
  if(p2 != NULL) {
    *p2='\0';
    p2++;

    p3=strchr(p2,'\n');
    if(p3 != NULL) {
      *p3='\0';
      p3++;

      p4=strchr(p3,'\n');
      if(p4 != NULL) {
        *p4='\0';
        p4++;

      }
    }
  }

  if(!strcasecmp(p2,"SMS_SEND")) {
    Serial.print("Sending SMS to ");
    Serial.print(p3);
    Serial.println("");
    Serial.println(p4);
    Serial.println("");
    SMSDev.SendSMSWithDisplay(p3, p4);
    
  } else if(!strcasecmp(p2,"DISPLAY")) {
    p5=strchr(p4,'\n');
    if(p5 != NULL) *p5='\0';
    Serial.print("Display:");
    Serial.print(p3);
    Serial.println(p4);
    SMSDev.Display(p3,p4);

  } else if(!strcasecmp(p2,"CLEAR")) {
    Serial.print("Clearing Display");
    SMSDev.Display("","");
    
  } else {
    Serial.println("(Ignored)");
    
  }
  Serial.println();
}

int SendTCPMessage(char *p1, char *p2, char *p3) {
  
  strcpy(TCPBuffer,WiGGIP);
  strcat(TCPBuffer,"\n");
  strcat(TCPBuffer,p1);
  if(p2 != NULL) {
    strcat(TCPBuffer,"\n");
    strcat(TCPBuffer,p2);
  }
  if(p3 != NULL) {
    strcat(TCPBuffer,"\n");
    strcat(TCPBuffer,p3);;
  }

  Serial.println();
  Serial.print("Message to: ");
  Serial.print(ServerIPAddr);
  Serial.print(" (Port: ");
  Serial.print(ServicePort);
  Serial.println(")");

  Serial.println(TCPBuffer);
 
  if (client.connect(ServerIPAddr,ServicePort)) {
    
    client.write(TCPBuffer);
//  client.write("\r\n");

    Serial.println("(Forwarded)");
    Serial.println();
    client.stop();
    return 1;
    
  } else {
    
    Serial.println("(Failed)");
    Serial.println();
    client.stop();
    return 0;
  }
}

// ---------------------------- Display Functions ---------------------------------------

void DisplayAllInfo() {

    SMSDev.DisplayTime();
    DisplayNetInfo();
    SMSDev.DisplayCfgInfo();
    SMSDev.AuthCallers = SMSDev.DisplayAuthCallers();
}

void DisplayNetInfo() {

  SMSDev.DisplayWithDelay("MAC Address", WiGGMAC, 1);

  SMSDev.DisplayWithDelay("IP Address", WiGGIP, 1);
  
  SMSDev.DisplayWithDelay("Server Address", ServerIP, 1);

  SMSDev.DisplayWithDelay("Service Port", itoa(ServicePort, Temp, 10), 1);
}

// ---------------------------- Button Functions ---------------------------------------

void SetButtons() {

  PCF.begin(); 
   
}

void CheckButtons() {

  if(PCF.read(0) == 0) {
    SMSDev.DisplayIdleScreen();
  }
  
  if(PCF.read(1) == 0) {
    DisplayAllInfo();
    SMSDev.DisplayIdleScreen();
  }
  
  if(PCF.read(2) == 0) {
    if (DisplayCountdown(2, "Requesting HELP")) {
      strcpy(Buffer, "ASSISTANCE REQUESTED!");
      if(strlen(SMSDev.DevName)) {
        strcat(Buffer, "\nfrom ");
        strcat(Buffer, SMSDev.DevName);
      }
      SMSDev.SendSMSWithDisplay(Notify, Buffer); 
    }
  }
  
  if(PCF.read(3) == 0) {
    if(DisplayCountdown(3, "Factory Reset")) {
      SMSDev.FactoryReset();
      CommandRESET();
    }
  }
  
}

int DisplayCountdown(int Button, char* Msg) {
unsigned long ResetTimer;
char ALine[20];
int SCheck;
int SecsRemaining = -1;
int SecsToWait = 5;

  ResetTimer = millis();
  do {
    SCheck = SecsToWait - ((millis() - ResetTimer) / 1000);
    if(SecsRemaining != SCheck) {
      SecsRemaining = SCheck;
      sprintf(ALine,"in %d Seconds...", SecsRemaining);
      SMSDev.Display(Msg, ALine);
    }
    if(SecsRemaining == 0) return 1;
    
  } while (PCF.read(Button) == 0);

  SMSDev.DisplayIdleScreen();

  return 0;
}

// ---------------------------- Command Processing Functions ---------------------------------------

void CommandRESET() {
  // Majority of Factory Reset handled by SMSDEV (inc Notification)
  // Do any Local Resets here...
  SetDefaults();
}

void CommandHELP() {
  
  strcpy(Buffer, DEVICE);
  strcat(Buffer, " Commands:\n\n");
  
  strcat(Buffer, "STATUS\n\n");
  strcat(Buffer, "(Other Config CMDs require Auth)\n\n");

  SMSDev.SendSMSWithDisplay(Caller,Buffer);  
}

void CommandSTATUS(char* acaller, char *msg) {
char csqval[2];

  strcpy(Buffer,"");

  if(msg != NULL) {
    strcpy(Buffer, msg);
    strcat(Buffer, "\n\n");
  }

  strcat(Buffer, DEVICE);
  strcat(Buffer, ": ");
  strcat(Buffer, VERSION);
  strcat(Buffer, "\n");
    
  strcat(Buffer, "Time: ");
  strcat(Buffer, SMSDev.TimeNow);
  strcat(Buffer, "\n");

  csqval[0]=char(SMSDev.CSQ + 48);
  csqval[1]='\0';
  strcat(Buffer, "Signal: ");
  strcat(Buffer, csqval);
  strcat(Buffer, "/5");
  strcat(Buffer, "\n");

  strcat(Buffer, "MAC: ");
  strcat(Buffer, WiGGMAC);
  strcat(Buffer, "\n");

  strcat(Buffer, "IPAddress: ");
  strcat(Buffer, WiGGIP);
  strcat(Buffer, "\n");

  strcat(Buffer, "Service Port: ");
  strcat(Buffer, itoa(ServicePort,Temp,10));
  strcat(Buffer, "\n");

  strcat(Buffer, "Server IP: ");
  strcat(Buffer, ServerIP);
  strcat(Buffer, "\n");

  SMSDev.SendSMSWithDisplay(Caller,Buffer);  
}

// ------------------------- Setup and Configuration --------------------------------------------

void SetConfig(char *Value,  char *Variable, int EEFieldNo, char *ReplyMsg, char *ReplyTo) {
  strncpy(Variable, Value, EEFIELDSIZE);
  Variable[EEFIELDSIZE-1] = '\0';
  SMSDev.WriteEEPROMString(EEFieldNo, Variable, EEFIELDSIZE);
  SMSDev.SendSMSWithDisplay(ReplyTo, ReplyMsg);  
}

int GetDefaults() {

  SMSDev.ReadEEPROMField(EEDEVSPEC1, WiFiSSID);
  if(strlen(WiFiSSID) == 0) return 0;

  SMSDev.ReadEEPROMField(EEDEVSPEC2, WiFiPwd);
  SMSDev.ReadEEPROMField(EEDEVSPEC3, ServerIP);
  SMSDev.ReadEEPROMField(EEPWD,      Password);
  SMSDev.ReadEEPROMField(EEDEVNAME,  DevName);
  SMSDev.ReadEEPROMField(EEDEVNUM,   DevNum);
  
  SMSDev.GetAuthCaller(0, Notify);  
  return 1;
}

void SetDefaults() {

// OPTIONAL:  May be set via SMS Command
  SiteDefaults();

  SMSDev.WriteEEPROMField(EEDEVSPEC1, WiFiSSID);
  SMSDev.WriteEEPROMField(EEDEVSPEC2, WiFiPwd);
  SMSDev.WriteEEPROMField(EEDEVSPEC3, ServerIP);
  SMSDev.WriteEEPROMField(EEPWD,      Password);
  SMSDev.WriteEEPROMField(EEDEVNAME,  DevName);
  SMSDev.WriteEEPROMField(EEDEVNUM,   DevNum);

  SMSDev.Reboot();
}

// ----------------------------- MAC / IP Conversion Functions -----------------------------------------------------

int MACFromString(char *macstr, byte *macaddr) {
int i;
char swork[18];
char *p1;
char *p2;

  if(CheckMACString(macstr)) return 0;

  strcpy(swork,macstr);
  strupr(swork);
  
  p1=swork;
  for(i=0; i<5; i++){
    p2=strchr(p1,':');
   *p2='\0';
    macaddr[i]=strtol(p1,NULL,16);
    p1=p2+1;
  }
  macaddr[5]=strtol(p1,NULL,16);
  
  return 1;
}

int MACToString(byte *macaddr, char *macstr) {

  sprintf(macstr, "%02X:%02X:%02X:%02X:%02X:%02X",
    macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4], macaddr[5]); 

  return 1;
}

int IPFromString(char *ipstr, IPAddress *ipaddr) {
int i;
char swork[16];
char *p1;
char *p2;

  if(CheckIPString(ipstr)) return 0;

  strcpy(swork,ipstr);
  p1=swork;
  for(i=0; i<3; i++){
    p2=strchr(p1,'.');
   *p2='\0';
    (*ipaddr)[i]=atoi(p1);
    p1=p2+1;
  }
  (*ipaddr)[3]=atoi(p1);
  
  return 1;
}

int IPFromString(char *ipstr, byte *ipaddr) {
int i;
char swork[16];
char *p1;
char *p2;

  if(CheckIPString(ipstr)) return 0;

  strcpy(swork,ipstr);
  p1=swork;
  for(i=0; i<3; i++){
    p2=strchr(p1,'.');
   *p2='\0';
    ipaddr[i]=atoi(p1);
    p1=p2+1;
  }
  ipaddr[3]=atoi(p1);
  
  return 1;
}

int CheckMACString(char *macstr) {
int i;
int c;
int slen;
char swork[18];

  slen=strlen(macstr);
  if(slen != 17) return 3;

  strcpy(swork,macstr);
  strupr(swork);
  
  for(i=0; i<slen; i++){
    c=(int)swork[i];
    if(c < 48) return 1;
    if((c > 58) && (c < 65)) return 1;
    if(c > 70) return 1;
  }

  for(i=2; i<slen; i=i+3){
    if(swork[i] != ':') return 2;
  }

  return 0;
}

int CheckIPString(char *ipstr) {
int i;
int slen;
int pcount;
int dcount;

  pcount=0;
  dcount=0;
  slen=strlen(ipstr);
  for(i=0; i<slen; i++){
    if(isDigit(ipstr[i])) dcount++;
    if(ipstr[i] == '.') pcount++;
  }
  if(pcount != 3) return 1;
  if((pcount+dcount) != slen) return 2;
  if((pcount+dcount) > 15) return 3;
  if((pcount+dcount) < 7) return 4;

  return 0;
}
